/*
  Copyright 2003, 2004 Julian Stecklina
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. The name of the author may not be used to endorse or promote products 
     derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*/

/* 
   $Id: plugin.c,v 1.17 2004/09/17 01:30:52 boelthorn Exp $
   This is the C part of the ECL X-Chat plugin. It defines various wrappers
   to call X-Chat's routines from Common Lisp scripts.
*/

#include "xchat-plugin.h"
#include <ecl.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>

#define ECL_PLUGIN_VERSION "0.2"

static xchat_plugin *ph;

/* 
   FIXME: 
   keep_ and remove_reference are used to keep references of objects X-Chat holds
   in memory the ECL garbage collector sees, so it will not be freed by mistake.
   Someone with a good insight into this may please suggest a better alternative.
*/

static void keep_reference(cl_object cons)
{
	cl_funcall(2,cl_intern(2,make_simple_string("KEEP-REFERENCE"),
			       make_simple_string("XCHAT")), cons);
}

static void remove_reference(cl_object cons)
{
	cl_funcall(2,cl_intern(2,make_simple_string("REMOVE-REFERENCE"),
			       make_simple_string("XCHAT")), cons);
}


/* Interns a C string */
static inline cl_object str_intern(char *str)
{
	return cl_intern(1,make_simple_string(str));
}

/* Coerce a C int to a Lisp number with regard to limited fixnum size */
static inline cl_object int2clnum(int d)
{
	if (d <= MOST_POSITIVE_FIXNUM && d >= MOST_NEGATIVE_FIXNUM)
                return MAKE_FIXNUM(d);
        else {
                cl_object x = big_register0_get();
                mpz_set_d(x->big.big_num, d);
                return big_register_copy(x);
        }
}

/* Coerce a Lisp number into a C int */
static inline int clnum2int(cl_object d)
{
	assert_type_integer(d);

	if (FIXNUMP(d))
		return fixint(d);
	else
		return (int)big_to_long(d);			  
}

/* Make a upcase keyword */
static inline cl_object make_upcase_keyword(const char *str)
{
	int len = strlen(str);
	char *ustr = calloc(len+1,sizeof(char));
	int i;
	cl_object keyword;

	for (i=0;i<=len;i++)
		ustr[i] = toupper(str[i]);
	ustr[len] = 0;
		
	keyword = make_keyword(ustr);
	/* Don't free ustr?!?! FIXME! */
	return keyword;
}

/* Constructs a list of strings out of a C array of strings */
static cl_object word2list(char **word)
{
	cl_object list;
	int i = 1;

	for (list = Cnil; (i <= 31) && word[i]!=NULL && word[i][0]!=0 ; i++) {
		list = CONS(make_string_copy(word[i]),list);
	}

	return cl_nreverse(list);
}

/* Get a copy of the name of a symbol */
static inline char *get_string_copy(cl_object str)
{
	char *lisp_str = str->string.self;
	char *c_str = calloc(strlen(lisp_str)+1,sizeof(char));

	strcpy(c_str,lisp_str); 
	return c_str;
}

/* 
 Lisp-callable wrappers for xchat functions 
*/

static cl_object cl_xchat_print(cl_object str)
{
	assert_type_string(str);
	xchat_print(ph,str->string.self);
	return Cnil;
}

static cl_object cl_xchat_command(cl_object str)
{
	assert_type_string(str);
	xchat_command(ph,str->string.self);
	return Cnil;
}

static cl_object cl_xchat_get_info(cl_object id)
{
	const char *info;
	assert_type_string(id);
	
	info = xchat_get_info(ph,id->string.self);
	
	if (info)
		return make_string_copy(info);
	else
		return Cnil;

}

static cl_object cl_xchat_get_prefs(cl_object name)
{
	int val;
	const char *str;

	assert_type_string(name);

	switch (xchat_get_prefs(ph,name->string.self,&str,&val))
	{
	case 0:	    /* Request failed */
		VALUES(0) = Cnil;
		VALUES(1) = Cnil;
		break;
	case 1:	    /* Got a string */
		VALUES(0) = make_string_copy(str);
		VALUES(1) = Ct;
		break;
	case 2:	    /* Got an integer */
		VALUES(0) = int2clnum(val);
		VALUES(1) = Ct;
		break;
	case 3:	    /* Got a boolean */
		if (val)
			VALUES(0) = Ct;
		else
			VALUES(0) = Cnil;
		VALUES(1) = Ct;
		break;
	default:    /* Got rubbish */
		FEerror("Unexpected type id from xchat.",0);
	}

		    /* Returns two values: the value of the x-xchat setting 
		       and a boolean denoting its existence */
        NVALUES = 2;
        return VALUES(0);
}

static cl_object cl_xchat_get_context()
{
	return int2clnum((int)xchat_get_context(ph)); /* FIXME: Bad representation? */
}

static cl_object cl_xchat_set_context(cl_object ctx)
{
	int result;

	assert_type_integer(ctx);

	result = xchat_set_context(ph,(xchat_context *)clnum2int(ctx));

	if (result==0)
		return Cnil;
	else 
		return Ct;

}

static cl_object cl_xchat_find_context(cl_object server, cl_object channel)
{
	xchat_context *ctx;
	char *serv;
	char *chan;

	if (server==Cnil)
		serv = NULL;
	else {
		assert_type_string(server);
		serv = server->string.self;
	}

	if (channel==Cnil)
		chan = NULL;
	else {
		assert_type_string(channel);
		chan = channel->string.self;
	}


	ctx = xchat_find_context(ph, serv, chan);
	if (ctx)
		return int2clnum((int)ctx);
	else
		return Cnil;
}


static cl_object cl_xchat_nickcmp(cl_object s1, cl_object s2)
{
	int result;

	assert_type_string(s1);
	assert_type_string(s2);

	result = xchat_nickcmp(ph,s1->string.self,s2->string.self);

		    /* Do some "normalizing" */
	if (result<0)
		return MAKE_FIXNUM(-1);
	else if (result==0)
		return MAKE_FIXNUM(0);
	else return MAKE_FIXNUM(1);
}

/*
  Callbacks have their corresponding lisp function as userdata, as the 
  need for such 'userdata' is not given in a language with lexical
  closures.
*/

static int cbGenericCommand(char *word[], char *word_eol[], void *userdata)
{
	cl_object result;
	cl_object data = (cl_object)userdata;
	cl_object lisp_callback = data;
	
	result = cl_funcall(4, cl_intern(2,make_simple_string("XCHAT-SAFE-FUNCALL"),
					 make_simple_string("XCHAT")),
			    lisp_callback,word2list(word),word2list(word_eol));

	if ( (type_of(result) == t_fixnum) || (type_of(result) == t_bignum) )
		return clnum2int(result);
	else
		return XCHAT_EAT_NONE; /* Discard return value, if no number */
}

static int cbGenericPrint(char *word[], void *userdata)
{
	cl_object result;
	cl_object data = (cl_object)userdata;
	cl_object lisp_callback = data;

	result = cl_funcall(3, cl_intern(2,make_simple_string("XCHAT-SAFE-FUNCALL"), 
 					 make_simple_string("XCHAT")), 
					 lisp_callback,word2list(word));  

	if ( (type_of(result) == t_fixnum) || (type_of(result) == t_bignum) )
		return clnum2int(result);
	else
		return XCHAT_EAT_NONE; /* Discard return value, if no number */
}


static int cbGenericTime(void *userdata)
{
	cl_object result;
	cl_object data = (cl_object)userdata;
	cl_object lisp_callback = data;
	
	result = cl_funcall(2, cl_intern(2,make_simple_string("XCHAT-SAFE-FUNCALL"),
					 make_simple_string("XCHAT")),
			    lisp_callback);

	if (Null(result)) {
		remove_reference(lisp_callback);
		return 0;
	} else
		return 1;	/* Keep timer going if non-nil */
}

/* 
   Lisp-callable wrappers for X-Chat's hooking facilities.
*/

static cl_object cl_xchat_hook_print(cl_object name, cl_object pri, cl_object callb)
{

	assert_type_string(name);
	assert_type_integer(pri);

	keep_reference(callb);

	return int2clnum((int)xchat_hook_print(ph,name->string.self,clnum2int(pri),
					       cbGenericPrint,callb));
}

static cl_object cl_xchat_hook_command(cl_object name, cl_object pri, cl_object callb,
				       cl_object help_text)
{
	assert_type_string(name);
	assert_type_integer(pri);
	assert_type_string(help_text);

	keep_reference(callb);

	return int2clnum((int)xchat_hook_command(ph,name->string.self,clnum2int(pri),cbGenericCommand,
						 help_text->string.self,callb));
}

static cl_object cl_xchat_hook_server(cl_object name, cl_object pri, cl_object callb)
{
	assert_type_string(name);
	assert_type_integer(pri);

	keep_reference(callb);

	return int2clnum((int)xchat_hook_server(ph,name->string.self,clnum2int(pri),cbGenericCommand,callb));
}

static cl_object cl_xchat_hook_timer(cl_object timeout, cl_object callb)
{
	assert_type_integer(timeout);

	keep_reference(callb);

	return int2clnum((int)xchat_hook_timer(ph,clnum2int(timeout),cbGenericTime,callb));
}

static cl_object cl_xchat_unhook(cl_object hook)
{
	cl_object callb = (cl_object)xchat_unhook(ph,(xchat_hook *)clnum2int(hook));

	remove_reference(callb);

	return Cnil;
}

/* This returns the element names of a list and type information */
static cl_object cl_xchat_list_fields(cl_object listname)
{
	const char **list;

	assert_type_string(listname);

	list = xchat_list_fields(ph,listname->string.self);

	if (list) {
		return word2list((char **)list);
	} else {
		return Cnil;
	}	
}

/* Returns a xchat list as alist */
static cl_object cl_xchat_list_get(cl_object name)
{
	const char **typeinfo;
	int ty_pos;
	cl_object list = Cnil;
	xchat_list *clist;

	assert_type_string(name);

	typeinfo = xchat_list_fields(ph,name->string.self);
	clist    = xchat_list_get(ph,name->string.self);

	if (typeinfo && clist)
	{
		cl_object desc = Cnil;

		while (xchat_list_next(ph,clist))
		{
			cl_object content = Cnil;
			char type;
			const char *elname;

			ty_pos = 0;

			desc = Cnil;
			while (typeinfo[ty_pos]) {

				content = Cnil;
				type = typeinfo[ty_pos][0];
				elname = typeinfo[ty_pos]+1;

				switch (type) {
				case 's':
				{ 
				  char *str = xchat_list_str(ph,clist,elname);
					cl_object obj;

					if (str)
						obj=make_string_copy(str);
					else
						obj=Cnil;

					content = CONS(make_upcase_keyword(elname),
						       obj);
				}
					break;
				case 'p':
				case 'i':
					content = CONS(make_upcase_keyword(elname),
						       int2clnum(xchat_list_int(ph,clist,elname)));
					break;
				default:
					printf("Did not recognise %s\n",typeinfo[ty_pos]);
					break;
				}
				
				desc = CONS(content,desc);
				ty_pos++;
			}

			list = CONS(desc,list);
		}

		return list;
	} else {
		return Cnil;
	}
}

/* Returns a list of alists for every user in the current channel */
static cl_object cl_xchat_get_users() /* FIXME: Remove when list_* functions are working */
{
	cl_object list = Cnil;
	xchat_list *clist;

	clist = xchat_list_get(ph, "users");
	if (clist) {
		while (xchat_list_next(ph,clist)) {
			cl_object user_desc = Cnil;
			
			user_desc = CONS(CONS(make_keyword("NICK"),
					      make_string_copy(xchat_list_str(ph,clist,"nick")))
					 ,user_desc);
			user_desc = CONS(CONS(make_keyword("HOST"),
					      make_string_copy(xchat_list_str(ph,clist,"host")))
					 ,user_desc);
			user_desc = CONS(CONS(make_keyword("PREFIX"),
					      make_string_copy(xchat_list_str(ph,clist,"prefix")))
					 ,user_desc);

			list = CONS(user_desc,list);			
		}
		xchat_list_free(ph, clist);		
	} 
	
	return list;
}

/* Add an entry in X-Chat's plugin dialog */
static cl_object cl_xchat_plugingui_add(cl_object filename, cl_object name, cl_object desc,
					cl_object version)
{
	assert_type_string(filename);
	assert_type_string(name);
	assert_type_string(desc);
	assert_type_string(version);

	return int2clnum((int)xchat_plugingui_add(ph,get_string_copy(filename),get_string_copy(name),
						  get_string_copy(desc),get_string_copy(version),NULL));
}

/* Remove an entry in X-Chat's plugin dialog */
static cl_object cl_xchat_plugingui_remove(cl_object handle)
{
	xchat_plugingui_remove(ph,(void *)clnum2int(handle));
	return Cnil;
}

/* Emit a print event */
static cl_object cl_xchat_emit_print(int nargs, cl_object name, ...) /* FIXME: de-uglify */
{
	va_list ap;
	int i,res;
	char *arg[8];

	assert_type_string(name);
	
	if (nargs>8)
		FEerror("Too many arguments given to xchat-emit-print.",0);

       	for (i=0;i<8;i++)
		arg[i]=NULL;
	
	va_start(ap,name);

	for (i=0;i<nargs-1;i++) {
		cl_object obj = va_arg(ap,cl_object);

		assert_type_string(obj);
		arg[i]=obj->string.self;
	}
	
	res = xchat_emit_print(ph,name->string.self,
			       arg[0],arg[1],arg[2],arg[3],
			       arg[4],arg[5],arg[6],arg[7]);
	if (res)
		return Ct;
	else
		return Cnil;
}

void init_XCHAT_SYSTEM(cl_object flags);

/*
  Functions X-Chat will call. Our door to the world.
*/

int xchat_plugin_init(xchat_plugin *plugin_handle,
                      char **plugin_name,
                      char **plugin_desc,
                      char **plugin_version,
                      char *arg)
{
	cl_object old_package;
	cl_object sym_package;

	ph = plugin_handle;
	
	*plugin_name = "ECL";
	*plugin_desc = "Embeddable Common-Lisp scripting interface";
	*plugin_version = ECL_PLUGIN_VERSION;

	cl_boot(1,plugin_name); /* FIXME? */
	
				/* Set up XCHAT package */

	sym_package = str_intern("*PACKAGE*");
	old_package = SYM_VAL(sym_package);
	SYM_VAL(sym_package) = make_package(make_simple_string("XCHAT"),Cnil,Cnil);

				/* Register wrapper functions, keeping %XCHAT- prefix for functions not
				   supposed to be called by user code */

	cl_def_c_function(str_intern("%PRINT"),cl_xchat_print, 1);
	cl_def_c_function(str_intern("%COMMAND"),cl_xchat_command, 1);
	cl_def_c_function(str_intern("%GET-INFO"),cl_xchat_get_info, 1);
	cl_def_c_function(str_intern("%GET-PREFS"),cl_xchat_get_prefs, 1);
	cl_def_c_function(str_intern("%GET-CONTEXT"),cl_xchat_get_context, 0);
	cl_def_c_function(str_intern("%SET-CONTEXT"),cl_xchat_set_context, 1);
	cl_def_c_function(str_intern("%FIND-CONTEXT"),cl_xchat_find_context, 2);
	cl_def_c_function(str_intern("%NICKCMP"),cl_xchat_nickcmp, 2);
	cl_def_c_function(str_intern("%XCHAT-HOOK-COMMAND"),cl_xchat_hook_command, 4);
	cl_def_c_function(str_intern("%XCHAT-HOOK-PRINT"),cl_xchat_hook_print, 3);
	cl_def_c_function(str_intern("%XCHAT-HOOK-SERVER"),cl_xchat_hook_server, 3);
	cl_def_c_function(str_intern("%XCHAT-HOOK-TIMER"),cl_xchat_hook_timer, 2);
	cl_def_c_function(str_intern("%XCHAT-UNHOOK"),cl_xchat_unhook,1);
	cl_def_c_function(str_intern("%GET-USERS"),cl_xchat_get_users,0);
	cl_def_c_function(str_intern("%XCHAT-PLUGINGUI-ADD"),cl_xchat_plugingui_add,4);
	cl_def_c_function(str_intern("%XCHAT-PLUGINGUI-REMOVE"),cl_xchat_plugingui_remove,1);
	cl_def_c_function(str_intern("%LIST-FIELDS"),cl_xchat_list_fields,1);
	cl_def_c_function(str_intern("%XCHAT-LIST-GET"),cl_xchat_list_get,1);

	cl_def_c_function_va(str_intern("%EMIT-PRINT"),cl_xchat_emit_print);

	SYM_VAL(sym_package) = old_package; /* Restore package */

				/* Do the exciting stuff from inside lisp.
				   FIXME: Would be nice to have this documented... */

	read_VV(OBJNULL,init_XCHAT_SYSTEM);

	xchat_print(ph, "Common Lisp interface loaded\n");
	
	return 1;
}

int xchat_plugin_deinit()
{

/*
  FIXME: The plugin cannot be unloaded, as there is no way to free all the memory
         ECL allocated.

 	cl_funcall(1,cl_intern(2,make_simple_string("%XCHAT-SHUTDOWN"),make_simple_string("XCHAT"))); 
 	xchat_print(ph, "ECL plugin unloaded successfully!\n"); 

	return 1;
*/

	xchat_print(ph, "ECL plugin unloading is currently NOT implemented.");
	return 0;
}
