; Copyright 2003, 2004 Julian Stecklina
; All rights reserved.

; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions
; are met:
; 1. Redistributions of source code must retain the above copyright
;    notice, this list of conditions and the following disclaimer.
; 2. Redistributions in binary form must reproduce the above copyright
;    notice, this list of conditions and the following disclaimer in the
;    documentation and/or other materials provided with the distribution.
; 3. The name of the author may not be used to endorse or promote products 
;    derived from this software without specific prior written permission.

; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ``AS IS'' AND
; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER BE LIABLE
; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
; SUCH DAMAGE.

;; $Id: xchat-system.lisp,v 1.15 2004/09/17 01:30:52 boelthorn Exp $
;; This is the Common Lisp part of the ECL X-Chat plugin. Here we separate
;; all the scripts into their own package and keep track about hooks.

;; Final package setup

(defpackage "XCHAT"
    (:use "COMMON-LISP")
  (:export "%PRINT" "%COMMAND" "%GET-INFO"
	   "%GET-PREFS" "%GET-CONTEXT" "%SET-CONTEXT"
	   "%FIND-CONTEXT" "%NICKCMP" "%HOOK-COMMAND"
	   "%HOOK-SERVER" "%HOOK-TIMER" "%UNHOOK"
	   "%PRINTF" "%COMMANDF" "%HOOK-PRINT"
	   "%GET-USERS" "%HOOK-CHANNEL-COMMAND"
	   "%XCHAT-PLUGIN-REGISTER" "%XCHAT-PLUGIN-AUTOLOAD"
	   "%LIST-FIELDS" "%LIST-GET"
	   "*XCHAT-AUTOLOAD-PLUGINS*" "*XCHAT-LISP-DIR*"
	   "DEF-XCHAT-PLUGIN" "%REGISTER-DEINIT-FUN"
	   "+XCHAT-PRI-NORM+" "+XCHAT-EAT-NONE+"
	   "+XCHAT-EAT-XCHAT+" "+XCHAT-EAT-PLUGIN+"
	   "+XCHAT-EAT-ALL+")
  (:documentation "A package exporting all you need to access X-Chat"))

(in-package "XCHAT")

;; Some generic macros

(defmacro bind-if ((arg test) then &optional else)
  "Bind arg to the result of test and execute 'then' or 'else'"
  (let ((tt (gensym)))
    `(let ((,tt ,test))
      (if ,tt
	  (let ((,arg ,tt))
	    ,then)
	  ,else))))
  
(defmacro bind-when ((arg test) &body body)
  `(bind-if (,arg ,test)
    (progn ,@body)))

;; X-Chat constants

(defconstant +XCHAT-PRI-NORM+   0)
(defconstant +XCHAT-EAT-NONE+   0)
(defconstant +XCHAT-EAT-XCHAT+  1)
(defconstant +XCHAT-EAT-PLUGIN+ 2)
(defconstant +XCHAT-EAT-ALL+    3)

;; Variables

;; We need to append a / so it is interpreted as directory.
(defvar *xchat-lisp-dir* (format nil "~A/"(%get-info "xchatdir"))
  "X-Chat configuration directory. Usually ~/.xchat/ or ~/.xchat2/.")

;; Convenience functions

(defun %printf (fmt-str &rest format-args)
  "A better xchat-printf than supplied by plugin interface"
  (%print (apply #'format nil fmt-str format-args)))

(defun %commandf (fmt-str &rest format-args)
  "A better xchat-commandf than supplied by plugin interface"
  (%command (apply #'format nil fmt-str format-args)))

(defmacro xchat-safe-eval (&body body)
  "Evaluate body catching all errors"
  `(handler-case
    (progn ,@body)
    (error (error) (progn 
		     (format t "An error was signaled: ~A~%" error)
		     error))))
  
(defun xchat-safe-funcall (&rest funcall-args)
  "funcall replacement with error catching"
  (xchat-safe-eval
   (apply #'funcall funcall-args)))

(defun %list-get (list-name)
  "Returns the corresponding list or NIL"
  (let ((name (string-downcase (string list-name))))
    (%xchat-list-get name)))

;; Keeping references on objects X-Chat holds (FIXME: Really neccessary?)

(defparameter *references* nil
  "References of objects X-Chat holds")

(defun keep-reference (obj)
  "Save object from being garbage collected"
  (push obj *references*))

(defun remove-reference (obj)
  "Let obj be garbage collected"
  (setq *references* (delete obj *references*)))

;; Getting *standard-output* redirected to X-Chat.

(defclass xchat-stream
    ()
  ((lbuffer :accessor xchat-stream-buffer
	    :initform (make-array  0 :adjustable t :fill-pointer t
				   :element-type 'character))
   (dirty   :accessor dirty-p
	    :initform nil))
  (:documentation "A stream directed to X-Chat"))

(defmethod ext::stream-close ((stream xchat-stream))
  (error "Cannot close ~S." stream))

(defmethod ext::stream-input-p ((stream xchat-stream))
  nil)

(defmethod ext::stream-output-p ((stream xchat-stream))
  t)

(defmethod stream-flush-buffer ((stream xchat-stream))
  (if (dirty-p stream)
      (%print (xchat-stream-buffer stream)))
  (ext::stream-clear-output stream))
    
(defmethod ext::stream-write-char ((stream xchat-stream) char)
  (setf (dirty-p stream) t)
  (if (char= char #\Newline)
      (stream-flush-buffer stream)
      (vector-push-extend char (xchat-stream-buffer stream))))

(defmethod ext::stream-force-output ((stream xchat-stream))
  (declare (ignore stream))
  nil)

(defmethod ext::stream-clear-output ((stream xchat-stream))
  (setf (xchat-stream-buffer stream) (make-array  0 :adjustable t :fill-pointer t
   						  :element-type 'character)
	(dirty-p stream) nil))

(setq *standard-output* (make-instance 'xchat-stream))

;; Basic plugin handling

(defvar *registered-plugins* (make-hash-table :test 'equal)
  "We keep the plugin structures here")

(defvar *xchat-autoload-plugins* nil
  "Use the config file to push files onto this list")

(defstruct plugin
  "Plugin description structure"
  name 
  deinit 
  package 
  handle 
  file
  (hooks nil))

;; Wrapping hooks into a class hierarchy

(defclass hook ()
  ((plugin :accessor hook-plugin
	   :initarg :plugin))
  (:documentation "A base class for all types of hooks"))

(defclass xchat-hook (hook)
  ((handle :accessor hook-handle
	   :initarg :handle))
  (:documentation "A class to hide X-Chat internals regarding hooks"))

;; If a plugin creates a hook note it in its plugin-object
(defmethod initialize-instance :after ((instance hook) &rest initargs &key &allow-other-keys)
  (push instance (plugin-hooks (hook-plugin instance))))

;; Remove this reference if it is no longer needed
(defmethod %unhook ((hook hook))
  (setf (plugin-hooks (hook-plugin hook))
	(delete hook (plugin-hooks (hook-plugin hook))))
  nil)

(defun %xchat-plugin-register (name version description &optional deinit package)
  "Registers a plugin in the X-Chat GUI dialog thingy. Plugins have to call this when they
   are loaded!"
  (cond
    ((not (boundp '*file-name*))
     (format t "~A tries to register itself, but not when loaded!~%" name)
     nil)
    (t (let ((handle (%xchat-plugingui-add (if (pathnamep *file-name*)
					       (format nil "~A" *file-name*)
					       *file-name*)
					   name description version)))
	 (setf (gethash name *registered-plugins*)
	       (make-plugin :name name
			    :file *file-name*
			    :handle handle
			    :deinit deinit
			    :package package))))))

(defun xchat-plugin-find (filename)
  "Find a plugin structure by filename"
  (maphash #'(lambda (name plugin)
	       (if (and (string-equal (pathname-name filename)
				      (pathname-name (plugin-file plugin)))
			(string-equal (pathname-type filename)
				      (pathname-type (plugin-file plugin))))
		   (return-from xchat-plugin-find
		     plugin)))
	   *registered-plugins*))
	   

(defun xchat-plugin-remove (name &optional (update-hash t))
  "Remove a plugin by name and update hashtable if asked to."
  (bind-when (plugin (or (gethash name *registered-plugins*)
			 (xchat-plugin-find name)))
	     (format t "Unloading ~A...~%" (plugin-name plugin))
	     (bind-when (deinit (plugin-deinit plugin))
			(funcall deinit))
	     (bind-when (package (plugin-package plugin))
			(delete-package package))
	     (loop for hook in (copy-list (plugin-hooks plugin)) ; Be careful here as %unhook modifies the list
		   do (%unhook hook))	
	     (%xchat-plugingui-remove (plugin-handle plugin))
	     (if update-hash
		 (remhash (plugin-name plugin) *registered-plugins*))
	     t))

(defun %xchat-shutdown ()
  "This gets called when X-Chat wants to unload the ECL plugin"
  (maphash
   #'(lambda (name plugin)
       (xchat-plugin-remove name nil)) *registered-plugins*)
  (clrhash *registered-plugins*))

(defun plugin-load (file)
  "Load a plugin from a file."
  (bind-if (filename (or (probe-file file)
			 (merge-pathnames file *xchat-lisp-dir*)))
	   (let ((*file-name* file))
	     (declare (special *file-name*))
	     (xchat-plugin-remove filename)
	     (format t "Loading ~A...~%" filename)
	     (load filename :verbose nil))
	   (format t "Could not find ~A!~%" file)))

(defun %xchat-plugin-autoload ()
  "Load all plugins specified in the config file"
  (handler-case
      (mapc #'plugin-load  *xchat-autoload-plugins*)
    (error ()
	   (format t "An error occured while reading the configuration file."))))

(defun cbLoad (word word_eol)
  "Hook to be called when X-Chat executes /load"
  (let ((file (second word)))
    (cond
      ((and file
	    (or (string-equal (pathname-type file) "LISP")
		(string-equal (pathname-type file) "LSP")))
       (plugin-load file)
       +XCHAT-EAT-ALL+)
      (t +XCHAT-EAT-NONE+))))

(defun cbUnload (word word_eol)
  "Hook to be called when X-Chat executes /unload"
  (let ((file (second word)))
    (if (and file
	     (xchat-plugin-remove file))
	+XCHAT-EAT-ALL+
	+XCHAT-EAT-NONE+)))

(defun new-hook (plugin handle)
  "Create a hook instance"
  (make-instance 'xchat-hook :handle handle
		             :plugin plugin))

(defun %hook-print (plugin name callb)
  "Register a print hook"
  (new-hook plugin (%xchat-hook-print name +XCHAT-PRI-NORM+ callb)))

(defun %hook-command (plugin name callb &optional (help-text ""))
  "Register a command hook"
  (new-hook plugin (%xchat-hook-command name +XCHAT-PRI-NORM+ callb
					help-text)))

(defun %hook-server (plugin name callb)
  "Register a raw server hook"
  (new-hook plugin (%xchat-hook-server name +XCHAT-PRI-NORM+ callb)))

(defun %hook-timer (plugin timeout callb)
  "Register a timer hook"
  (let ((hook (new-hook plugin nil)))
    (setf (hook-handle hook) (%xchat-hook-timer timeout 
						#'(lambda ()
						    (unless (funcall callb)
						      (%unhook hook))
						    t)))))

;; If a hook is a normal X-Chat hook, tell X-Chat to unhook it
(defmethod %unhook :after ((hook xchat-hook))
  (%xchat-unhook (hook-handle hook)))


(defmacro def-xchat-plugin (name &key (version "") (description "")
				      (use-packages '("COMMON-LISP" "XCHAT"))
				      (handle-var '*plugin*)
				      (package-name name))
  "Create a package for the plugin and register it."
  (let ((package (gensym)))
    `(let ((,package (make-package ,package-name :use (quote ,use-packages))))
      (in-package ,package-name)
      (set (intern ,(string handle-var) ,package)
       (%xchat-plugin-register ,name ,version ,description nil ,package)))))
    
(defun %register-deinit-fun (plugin fn)
  "Register a function that gets called, when the plugin is unloaded"
  (setf (plugin-deinit plugin) fn))

(%xchat-hook-command "load" +XCHAT-PRI-NORM+ #'cbLoad
		     "")

(%xchat-hook-command "unload" +XCHAT-PRI-NORM+ #'cbUnload
		     "")

;; Showing off by implementing !COMMAND's
;;
;; This is just intended as example of how to extend hooks

(defvar *registered-commands* (make-hash-table :test 'equal))

(defclass channel-hook (hook)
  ((cmd :accessor hook-cmd
	:initarg :cmd)
   (callb :accessor hook-callback
	  :initarg :callback))
  (:documentation "Register a hook that reacts on !cmd in channels"))

(defmethod initialize-instance :after ((instance channel-hook) &rest initargs &key &allow-other-keys)
  (setf (gethash (hook-cmd instance) *registered-commands*) instance))

(defun %hook-channel-command (plugin cmd callb)
  (make-instance 'channel-hook :cmd (string-upcase cmd)
			     :plugin plugin
			     :callback callb))

(defmethod %unhook :after ((hook channel-hook))
  (remhash (hook-cmd hook) *registered-commands*))

(defun split-string (string &key (char #\Space) (start 0))
  "Split a string into pieces"
  (declare (base-char char)
	   (simple-string string)
	   (fixnum start))
  (bind-if (pos (position char string :start start))
     (let ((result (split-string string :char char :start (1+ pos))))
       (if (= start pos)
	   result
	 (cons (subseq string start pos)
	       result)))
     (if (/= start (length string))
	 (list (subseq string start)))))

(defun cbMessage (word)
  "The callback that distributes a specific command to the responsible function"
  (let ((split (split-string (second word))))
    (if (and (first split)
	     (not (zerop (length (first split)))))
	(bind-if (hook (gethash (string-upcase (first split)) *registered-commands*))
		 (funcall (hook-callback hook) split word))))
  +XCHAT-EAT-NONE+)

(%xchat-hook-print "Your Message" +XCHAT-PRI-NORM+ #'cbMessage)
(%xchat-hook-print "Channel Message" +XCHAT-PRI-NORM+ #'cbMessage)

;; Make X-Chat stuff available in USER package

(use-package "XCHAT" "COMMON-LISP-USER")

;; Load all .lisp files in ~/.xchat/

(loop for file in (directory (merge-pathnames 
			      (make-pathname :name :wild :type "lisp"
					     :case :local)
			      *xchat-lisp-dir*))
      do (plugin-load file))
